// Part of SimCoupe - A SAM Coupe emulator
//
// Display.cpp: SDL display rendering
//
//  Copyright (c) 1999-2006  Simon Owen
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or
// (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

// ToDo:
//  - change to handle multiple dirty regions

#include "SimCoupe.h"

#include "Display.h"
#include "Frame.h"
#include "GUI.h"
#include "Options.h"
#include "Profile.h"
#include "Video.h"
#include "UI.h"

#include "psp_kbd.h"
#include "psp_sdl.h"
#include "psp_danzeff.h"

bool* Display::pafDirty;
SDL_Rect rSource, rTarget;

////////////////////////////////////////////////////////////////////////////////

// Writing to the display in DWORDs makes it endian sensitive, so we need to cover both cases
#if SDL_BYTEORDER == SDL_LIL_ENDIAN

inline DWORD PaletteDWORD (BYTE b1_, BYTE b2_, BYTE b3_, BYTE b4_, DWORD* pulPalette_)
    { return (pulPalette_[b4_] << 24) | (pulPalette_[b3_] << 16) | (pulPalette_[b2_] << 8) | pulPalette_[b1_]; }
inline DWORD PaletteDWORD (BYTE b1_, BYTE b2_, DWORD* pulPalette_)
    { return (pulPalette_[b2_] << 16) | pulPalette_[b1_]; }
inline DWORD MakeDWORD (BYTE b1_, BYTE b2_, BYTE b3_, BYTE b4_)
    { return (b4_ << 24) | (b3_ << 16) | (b2_ << 8) | b1_; }
inline DWORD MakeDWORD (BYTE b1_, BYTE b2_)
    { return ((b2_ << 16) | b1_) * 0x0101UL; }
inline DWORD JoinWORDs (DWORD w1_, DWORD w2_)
    { return ((w2_ << 16) | w1_); }
#else

inline DWORD PaletteDWORD (BYTE b1_, BYTE b2_, BYTE b3_, BYTE b4_, DWORD* pulPalette_)
    { return (pulPalette_[b1_] << 24) | (pulPalette_[b2_] << 16) | (pulPalette_[b3_] << 8) | pulPalette_[b4_]; }
inline DWORD PaletteDWORD (BYTE b1_, BYTE b2_, DWORD* pulPalette_)
    { return (pulPalette_[b1_] << 16) | pulPalette_[b2_]; }
inline DWORD MakeDWORD (BYTE b1_, BYTE b2_, BYTE b3_, BYTE b4_)
    { return (b1_ << 24) | (b2_ << 16) | (b3_ << 8) | b4_; }
inline DWORD MakeDWORD (BYTE b1_, BYTE b2_)
    { return ((b1_ << 16) | b2_) * 0x0101UL; }
inline DWORD JoinWORDs (DWORD w1_, DWORD w2_)
    { return ((w1_ << 16) | w2_); }
#endif

////////////////////////////////////////////////////////////////////////////////

bool Display::Init (bool fFirstInit_/*=false*/)
{
    Exit(true);

    pafDirty = new bool[Frame::GetHeight()];

    // These will be updated to the appropriate values on the first draw
    rSource.w = rTarget.w = Frame::GetWidth();
    rSource.h = rTarget.h = Frame::GetHeight() << 1;

    return Video::Init(fFirstInit_);
}

void Display::Exit (bool fReInit_/*=false*/)
{
    Video::Exit(fReInit_);

    if (pafDirty) { delete[] pafDirty; pafDirty = NULL; }
}

void Display::SetDirty ()
{
    // Mark all display lines dirty
    for (int i = 0, nHeight = Frame::GetHeight() ; i < nHeight ; i++)
        pafDirty[i] = true;
}


// Draw the changed lines in the appropriate colour depth and hi/low resolution
bool DrawChanges (CScreen* pScreen_, SDL_Surface* pSurface_)
{
  // Lock the surface for direct access below
  if (SDL_MUSTLOCK(pSurface_) && SDL_LockSurface(pSurface_) < 0)
  {
    TRACE("!!! SDL_LockSurface failed: %s\n", SDL_GetError());
    return false;
  }

  //bool fInterlace = !GUI::IsActive();
  bool fInterlace = true;

  short *pdsStart = reinterpret_cast<short *>(pSurface_->pixels);
  short *pdsBack  = pdsStart;

  long lPitchDW = pSurface_->pitch >> (fInterlace ? 1 : 2);
  bool *pfDirty = Display::pafDirty, *pfHiRes = pScreen_->GetHiRes();

  BYTE *pbSAM = pScreen_->GetLine(0);
  BYTE *pb    = pbSAM;
  long lPitch = pScreen_->GetPitch();

  int nShift = fInterlace ? 1 : 0;
  int nDepth = pSurface_->format->BitsPerPixel;
  int nBottom = pScreen_->GetHeight() >> nShift;
  int nWidth = pScreen_->GetPitch(), nRightHi = nWidth, nRightLo = nRightHi >> 1;

  ProfileStart(Gfx);

  nWidth <<= 1;

  int y_d_p = 0;
  for (int y = 0 ; y < nBottom ; y++)
  {
    //LUDO: if (!pfDirty[y]) continue;

    int y_d = (y * 272) / nBottom;
    while (y_d_p <= y_d) 
    {
      pdsBack = pdsStart + (lPitchDW * y_d_p);
      short *pds = pdsBack;
  
      if (pfHiRes[y])
      {
        for (int x = 0 ; x < nRightHi; x++)
        {
          int x_d = (x * 480) / nRightHi;
          pds[x_d] = aulPalette[pb[x]];
        }
        if (fInterlace)
        {
          pb  = pbSAM;
          pds = (short *)(pdsBack + lPitchDW);
  
          if (!GetOption(scanlevel)) {
              memset(pds, 0x00, nWidth);
          }
          else
          {
            for (int x = 0 ; x < nRightHi; x++)
            {
              int x_d = (x * 480) / nRightHi;
              pds[x_d] = aulScanline[pb[x]];
            }
          }
        }
      }
      else
      {
        for (int x = 0 ; x < nRightLo; x++)
        {
          int x_d = (x * 480) / nRightLo;
          int color = aulPalette[pb[x]];
          pds[x_d  ] = color;
          pds[x_d+1] = color;
        }
    
        if (fInterlace)
        {
          pb = pbSAM;
          pds = (short *)(pdsBack + lPitchDW);
    
          if (!GetOption(scanlevel)) {
            memset(pds, 0x00, nWidth);
          }
          else
          {
            for (int x = 0 ; x < nRightLo; x++)
            {
              int x_d = (x * 480) / nRightLo;
              int color = aulScanline[pb[x]];
              pds[x_d  ] = color;
              pds[x_d+1] = color;
            }
          }
        }
      }
      y_d_p++;
    }

    pb  = (pbSAM   += lPitch);
  }

  // Unlock the surface now we're done drawing on it
  if (pSurface_ && SDL_MUSTLOCK(pSurface_))
      SDL_UnlockSurface(pSurface_);

  ProfileEnd();

  ProfileStart(Blt);


  // Calculate the source rectangle for the full visible area
  rSource.x = 0;
  rSource.y = 0;
  rSource.w = pScreen_->GetPitch();
  rSource.h = nBottom;

  // Calculate the target rectangle on the target display
  rTarget.x = ((screen_ptr->w - rSource.w) >> 1);
  rTarget.y = ((screen_ptr->h - (rSource.h << nShift)) >> 1);
  rTarget.w = rSource.w;
  rTarget.h = rSource.h << nShift;


  // Find the first changed display line
  int nChangeFrom = 0;
  for ( ; nChangeFrom < nBottom && !pfDirty[nChangeFrom] ; nChangeFrom++);

  if (nChangeFrom < nBottom)
  {
      // Find the last change display line
      int nChangeTo = nBottom-1;
      for ( ; nChangeTo && !pfDirty[nChangeTo] ; nChangeTo--);

      // Clear the dirty flags for the changed block
      for (int i = nChangeFrom ; i <= nChangeTo ; pfDirty[i++] = false);
  }

  ProfileEnd();

  // Success
  return true;
}


// Update the display to show anything that's changed since last time
void Display::Update (CScreen* pScreen_)
{
    // Don't draw if fullscreen but not active
    if (GetOption(fullscreen) && !g_fActive)
        return;

    // Draw any changed lines
    DrawChanges(pScreen_, screen_ptr);

    //LUDO:
    if (psp_kbd_is_danzeff_mode()) {
      danzeff_moveTo(-130, -50);
      danzeff_render();
    }

    psp_sdl_flip();

    if (psp_screenshot_mode) {
      psp_screenshot_mode--;
      if (psp_screenshot_mode <= 0) {
        psp_sdl_save_screenshot();
        psp_screenshot_mode = 0;
      }
    }
}


// Scale a client size/movement to one relative to the SAM view port size
// Should round down and be consistent with positive and negative values
void Display::DisplayToSamSize (int* pnX_, int* pnY_)
{
    int nHalfWidth = !GUI::IsActive(), nHalfHeight = 0;

    *pnX_ = *pnX_ * (rSource.w >> nHalfWidth)  / rTarget.w;
    *pnY_ = *pnY_ * (rSource.h >> nHalfHeight) / rTarget.h;
}

// Scale a size/movement in the SAM view port to one relative to the client
// Should round down and be consistent with positive and negative values
void Display::SamToDisplaySize (int* pnX_, int* pnY_)
{
    int nHalfWidth = !GUI::IsActive(), nHalfHeight = 0;

    *pnX_ = *pnX_ * rTarget.w / (rSource.w >> nHalfWidth);
    *pnY_ = *pnY_ * rTarget.h / (rSource.h >> nHalfHeight);
}

// Map a client point to one relative to the SAM view port
void Display::DisplayToSamPoint (int* pnX_, int* pnY_)
{
    *pnX_ -= rTarget.x;
    *pnY_ -= rTarget.y;
    DisplayToSamSize(pnX_, pnY_);
}

// Map a point in the SAM view port to a point relative to the client position
void Display::SamToDisplayPoint (int* pnX_, int* pnY_)
{
    SamToDisplaySize(pnX_, pnY_);
    *pnX_ += rTarget.x;
    *pnY_ += rTarget.y;
}
